package de.ozgcloud.eingang.xta;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;

class XtaServiceTest {

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private List<XtaMessageHandler> xtaMessageHandlers;

	@Mock
	private XtaRemoteService remoteService;

	@Mock
	private XtaIncomingFilesMapper incomingFilesMapper;

	@Mock
	private XtaMessageHandler messageHandler1;

	@Mock
	private XtaMessageHandler messageHandler2;

	@Mock
	private FormData formData;

	@DisplayName("get messages")
	@Nested
	class TestGetMessages {

		@Mock
		XtaMessageMetaData messageMetaData;

		@DisplayName("should return stream of FormData")
		@Test
		void shouldReturnStreamOfFormData() {
			doReturn(Stream.of(messageMetaData)).when(service).createXtaMessageStream();
			doReturn(Optional.of(formData)).when(service).mapToFormData(messageMetaData);

			var result = service.getMessages().toList();

			assertThat(result).containsExactly(formData);
		}

		@DisplayName("should return empty stream if no FormData optional")
		@Test
		void shouldReturnEmptyStreamIfNoFormDataOptional() {
			doReturn(Stream.of(messageMetaData)).when(service).createXtaMessageStream();
			doReturn(Optional.empty()).when(service).mapToFormData(messageMetaData);

			var result = service.getMessages().toList();

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("map to form data")
	@Nested
	class TestMapToFormData {

		@Mock
		private XtaMessageMetaData xtaMessageMetaData;

		@DisplayName("should return FormData from responsible handler")
		@Test
		void shouldReturnFormDataFromResponsibleHandler() {
			doReturn(Optional.of(messageHandler1)).when(service).getResponsibleHandler(xtaMessageMetaData);
			doReturn(formData).when(service).getFormData(xtaMessageMetaData, messageHandler1);

			var result = service.mapToFormData(xtaMessageMetaData);

			assertThat(result).contains(formData);
		}

		@DisplayName("should return empty if no responsible handler")
		@Test
		void shouldReturnEmptyIfNoResponsibleHandler() {
			doReturn(Optional.empty()).when(service).getResponsibleHandler(xtaMessageMetaData);

			var result = service.mapToFormData(xtaMessageMetaData);

			assertThat(result).isEmpty();
		}
	}

	@DisplayName("create XTA message stream")
	@Nested
	class TestCreateXtaMessageStream {

		@Mock
		private XtaMessageMetaDatasAndHeader messagesMetaData;

		@DisplayName("should return stream of XtaMessageMetaData")
		@Test
		void shouldReturnStreamOfXtaMessageMetaData() {
			var message = XtaMessageMetaDataTestFactory.create();
			when(remoteService.getMessagesMetadata()).thenReturn(messagesMetaData);
			when(messagesMetaData.getMessages()).thenReturn(Stream.of(message));

			var result = service.createXtaMessageStream().toList();

			assertThat(result).containsExactly(message);
		}
	}

	@DisplayName("get responsible handler")
	@Nested
	class TestGetResponsibleHandler {

		@Mock
		private XtaMessageMetaData messageMetaData;

		@DisplayName("should return first responsible handler")
		@Test
		void shouldReturnHandlerIfResponsible() {
			when(messageHandler1.isResponsible(messageMetaData)).thenReturn(true);
			when(xtaMessageHandlers.stream()).thenReturn(Stream.of(messageHandler1, messageHandler2));

			var result = service.getResponsibleHandler(messageMetaData);

			assertThat(result).contains(messageHandler1);
		}

		@DisplayName("with no handler responsible")
		@Nested
		class TestWithNoHandlerResponsible {
			private static final String UNKONWN_MESSAGE_TYPE = "UNKONWN_MESSAGE_TYPE";

			@BeforeEach
			void mock() {
				when(messageHandler1.isResponsible(messageMetaData)).thenReturn(false);
				when(messageHandler2.isResponsible(messageMetaData)).thenReturn(false);
				when(xtaMessageHandlers.stream()).thenReturn(Stream.of(messageHandler1, messageHandler2));
				when(messageMetaData.getMessageType()).thenReturn(UNKONWN_MESSAGE_TYPE);
				doNothing().when(service).logWarningForUnsupportedMessageType(UNKONWN_MESSAGE_TYPE);
			}

			@DisplayName("should return empty")
			@Test
			void shouldReturnEmptyIfNoHandlerResponsible() {
				var result = service.getResponsibleHandler(messageMetaData);

				assertThat(result).isEmpty();
			}

			@DisplayName("should log warning")
			@Test
			void shouldLogWarningIfNoHandlerResponsible() {
				service.getResponsibleHandler(messageMetaData);

				verify(service).logWarningForUnsupportedMessageType(UNKONWN_MESSAGE_TYPE);
			}
		}
	}

	@DisplayName("get form data")
	@Nested
	class TestGetFormData {

		@Mock
		private XtaMessage msg;

		@Mock
		private List<IncomingFile> incomingFiles;

		@Mock
		private List<XtaFile> xtaFiles;

		@DisplayName("should return form data")
		@Test
		void shouldReturnFormData() {
			var metaData = XtaMessageMetaDataTestFactory.create();
			when(remoteService.getMessage(metaData.getMessageId())).thenReturn(msg);
			when(msg.getMessageFiles()).thenReturn(xtaFiles);
			when(incomingFilesMapper.toIncomingFiles(xtaFiles)).thenReturn(incomingFiles);
			when(messageHandler1.getFormData(metaData, incomingFiles)).thenReturn(formData);

			var result = service.getFormData(metaData, messageHandler1);

			assertThat(result).isEqualTo(formData);
		}
	}

	@Nested
	class TestAcknowledgeReceive {

		@Test
		void shouldCallRemoteService() {
			service.acknowledgeReceive(XtaMessageTestFactory.MESSAGE_ID);

			verify(remoteService).close(XtaMessageTestFactory.MESSAGE_ID);
		}
	}

}
